import sys
import token
import tokenize

op_trans = { '+': token.PLUS, '-': token.MINUS, '*': token.STAR, '/':
             token.SLASH, '=': token.EQUAL, ';': token.SEMI, ':':
             token.COLON, '(': token.LPAR, ')': token.RPAR, ',':
             token.COMMA, }

keywords = { 'group': 1, 'types': 1, 'funcs': 1, 'consts': 1,
             'callbacks': 1, 'structs': 1, 'as': 1, }

class Token:
    def __init__(self, type, string, lineno):
        self._num = type
        if type == token.OP:
            type = op_trans[string]
        if type == token.NUMBER:
            self.value = int(string)
        else:
            self.value = string
        if type == token.NAME and string in keywords:
            self.type = string.upper()
        else:
            self.type = token.tok_name[type]
        self.lineno = lineno

    def __str__(self):
        return '<Token %d: %s(%d) %s>' % (self.lineno, self.type,
                                          self._num, repr(self.value))

class Lexer:
    def __init__(self, fp):
        self._stream = tokenize.generate_tokens(fp.readline)

    def token(self):
        type, tstr, (srow, scol), (erow, ecol), line = self._stream.next()
        if type == 53:                  # NL not in token.py?!?
            type = token.NEWLINE
        if type != token.ENDMARKER:
            return Token(type, tstr, srow)

tokens = ('AS', 'CALLBACKS', 'COMMENT', 'COLON', 'COMMA', 'CONSTS',
          'DEDENT', 'EQUAL', 'FUNCS', 'GROUP', 'INDENT', 'NAME',
          'NEWLINE', 'SLASH', 'STAR', 'STRING', 'STRUCTS', 'TYPES')

class Node:
    def __init__(self, **kw):
        self.__dict__ = kw

    def __repr__(self):
        # Crappy to XML transform for display
        parts = []
        parts.append('<node type="%s">\n' % self.type)
        keys = self.__dict__.keys()
        keys.sort()
        for name in keys:
            if name == 'type':
                continue
            content = getattr(self, name)
            if type(content) is type([]):
                parts.append('  <%s list="yes">\n' % name)
                for elem in content:
                    elemrep = str(elem)
                    parts.append('    ')
                    parts.append(elemrep.replace('\n', '\n    ').rstrip())
                    parts.append('\n')
                parts.append('  </%s>\n' % name)
            else:
                content = str(content)
                if content.count('\n') == 0:
                    parts.append('  <%s>%s</%s>\n' % (name, content, name))
                else:
                    parts.append('  <%s>\n' % name)
                    parts.append('    ')
                    parts.append(content.replace('\n', '\n    ').rstrip())
                    parts.append('\n')
                    parts.append('  </%s>\n' % name)
        parts.append('</node>\n')
        return ''.join(parts)

def uncomment(s):
    return s[1:].lstrip()

def unprefix(s):
    if s and s[0] == '\n':
        lines = s[1:].split('\n')
        strip = len(lines[0])
        for line in lines:
            line = line.rstrip()
            if len(line):
                cur = len(line) - len(line.lstrip())
                if cur < strip:
                    strip = cur
        new_lines = []
        for line in lines:
            new_lines.append(line[strip:])
        s = '\n'.join(new_lines)
    return s

def unstring(s):
    return unprefix(eval(s))

def p_module(t):
    'module : prog'
    t[0] = Node(type='prog', list=t[1])

# Collect a list of program elements
def p_prog_1(t):
    'prog : stmt'
    t[0] = []
    if t[1]:
        t[0].append(t[1])

def p_prog_2(t):
    'prog : prog stmt'
    if t[2]:
        t[1].append(t[2])
    t[0] = t[1]

# All statement variants
def p_stmt_0(t):
    'stmt : NEWLINE'
    pass

def p_stmt_1(t):
    '''stmt : group_stmt
            | const_stmt
            | callbacks_stmt
            | types_stmt
            | structs_stmt
            | funcs_stmt'''
    t[0] = t[1]

def p_stmt_2(t):
    'stmt : COMMENT'
    t[0] = Node(type='comment', text=uncomment(t[1]))

# New group of definitions
def p_group_stmt(t):
    'group_stmt : GROUP STRING NEWLINE'
    t[0] = Node(type='group', name=unstring(t[2]))

# List of constant definitions
def p_const_stmt(t):
    'const_stmt : CONSTS COLON NEWLINE INDENT const_list DEDENT'
    t[0] = Node(type='constlist', list=t[5])

def p_const_list_1(t):
    'const_list : const_def'
    t[0] = [t[1]]

def p_const_list_2(t):
    'const_list : const_list const_def'
    t[1].append(t[2])
    t[0] = t[1]

def p_const_def_1(t):
    'const_def : NAME NAME COLON NEWLINE INDENT value_list DEDENT'
    t[0] = Node(type='constgroup', c_type=t[1], option=None, group=t[2],
                list=t[6])

def p_const_def_2(t):
    'const_def : NAME NAME NAME COLON NEWLINE INDENT value_list DEDENT'
    t[0] = Node(type='constgroup', c_type=t[1], option=t[2], group=t[3],
                list=t[7])

def p_value_list_0(t):
    'value_list : value_list NEWLINE'
    t[0] = t[1]

def p_value_list_1(t):
    'value_list : NAME'
    t[0] = [t[1]]

def p_value_list_2(t):
    'value_list : value_list NAME'
    t[1].append(t[2])
    t[0] = t[1]

# List of types statement
def p_types_stmt(t):
    'types_stmt : TYPES COLON NEWLINE INDENT type_list DEDENT'
    t[0] = Node(type='typelist', list=t[5])

def p_type_list_0(t):
    'type_list : type_list NEWLINE'
    t[0] = t[1]

def p_type_list_1(t):
    'type_list : single_type'
    t[0] = [t[1]]

def p_type_list_2(t):
    'type_list : type_list single_type'
    t[1].append(t[2])
    t[0] = t[1]

def p_single_type_1(t):
    'single_type : basic_type EQUAL basic_type NEWLINE'
    t[0] = Node(type='typeequiv', name=t[1], equiv=t[3])

def p_single_type_2(t):
    'single_type : type_selector_list COLON NEWLINE INDENT type_param_list DEDENT'
    t[0] = Node(type='typedef', name=t[1], params=t[5])

def p_type_selector_list_1(t):
    'type_selector_list : type_selector'
    t[0] = [t[1]]

def p_type_selector_list_2(t):
    'type_selector_list : type_selector_list COMMA type_selector'
    t[1].append(t[3])
    t[0] = t[1]

def p_type_selector_1(t):
    'type_selector : basic_type'
    t[0] = Node(type='typesel', c_type=t[1])

def p_type_selector_2(t):
    'type_selector : basic_type type_opt'
    t[0] = Node(type='typesel', c_type=t[1], opt=t[2])

def p_type_selector_3(t):
    'type_selector : type_opt'
    t[0] = Node(type='typesel', opt=t[1])

def p_basic_type_1(t):
    'basic_type : NAME'
    t[0] = t[1]

def p_basic_type_2(t):
    'basic_type : NAME STAR'
    t[0] = t[1] + ' *'

def p_type_opt_1(t):
    'type_opt : SLASH NAME'
    t[0] = t[2]

def p_type_param_list_0(t):
    'type_param_list : type_param_list NEWLINE'
    t[0] = t[1]

def p_type_param_list_1(t):
    'type_param_list : type_param'
    t[0] = [t[1]]

def p_type_param_list_2(t):
    'type_param_list : type_param_list type_param'
    t[1].append(t[2])
    t[0] = t[1]

def p_type_param(t):
    'type_param : NAME STRING NEWLINE'
    t[0] = Node(type='typeparam', name=t[1], value=unstring(t[2]))

# List of structs statement
def p_structs_stmt(t):
    'structs_stmt : STRUCTS COLON NEWLINE INDENT struct_list DEDENT'
    t[0] = Node(type='structlist', list=t[5])

def p_struct_list_0(t):
    'struct_list : struct_list NEWLINE'
    t[0] = t[1]

def p_struct_list_1(t):
    'struct_list : single_struct'
    t[0] = [t[1]]

def p_struct_list_2(t):
    'struct_list : struct_list single_struct'
    t[1].append(t[2])
    t[0] = t[1]

def p_single_struct(t):
    'single_struct : NAME EQUAL NAME COLON NEWLINE INDENT member_list DEDENT'
    t[0] = Node(type='struct', name=t[1], inner=t[3], members=t[7])

def p_member_list_0(t):
    'member_list : member_list NEWLINE'
    t[0] = t[1]

def p_member_list_1(t):
    'member_list : single_member'
    t[0] = [t[1]]

def p_member_list_2(t):
    'member_list : member_list single_member'
    t[1].append(t[2])
    t[0] = t[1]

def p_single_member_1(t):
    'single_member : NAME NAME member_opt'
    t[0] = Node(type='member', c_type=t[1], name=t[2], rename=t[2], opt=t[3])

def p_single_member_2(t):
    'single_member : NAME NAME AS NAME member_opt'
    t[0] = Node(type='member', c_type=t[1], name=t[2], rename=t[4], opt=t[5])

def p_member_opt_0(t):
    'member_opt : '

def p_member_opt_1(t):
    'member_opt : COLON NEWLINE INDENT opt DEDENT'
    t[0] = t[4]

def p_opt_1(t):
    'opt : NAME newline_list'
    t[0] = Node(type='opt', name=t[1], value=None)

def p_opt_2(t):
    'opt : NAME STRING newline_list'
    t[0] = Node(type='opt', name=t[1], value=unstring(t[2]))

# List of funcs statement
def p_funcs_stmt(t):
    'funcs_stmt : FUNCS COLON NEWLINE INDENT func_list DEDENT'
    t[0] = Node(type='funclist', list=t[5])

def p_func_list_0(t):
    'func_list : func_list NEWLINE'
    t[0] = t[1]

def p_func_list_1(t):
    'func_list : single_func'
    t[0] = [t[1]]

def p_func_list_2(t):
    'func_list : func_list single_func'
    t[1].append(t[2])
    t[0] = t[1]

def p_single_func_1(t):
    'single_func : basic_type NAME NEWLINE'
    t[0] = Node(type='func', c_type=t[1], name=t[2], args=[], code=None)

def p_single_func_2(t):
    'single_func : basic_type NAME COLON NEWLINE INDENT func_args func_code DEDENT'
    t[0] = Node(type='func', c_type=t[1], name=t[2], args=t[6], code=t[7])

def p_func_args_0(t):
    'func_args : func_args NEWLINE'
    t[0] = t[1]

def p_func_args_1(t):
    'func_args : single_arg'
    t[0] = [t[1]]

def p_func_args_2(t):
    'func_args : func_args single_arg'
    t[1].append(t[2])
    t[0] = t[1]

def p_func_code_0(t):
    'func_code : '
    pass

def p_func_code_1(t):
    'func_code : STRING newline_list'
    t[0] = unstring(t[1])

def p_newline_list(t):
    '''newline_list : NEWLINE
                    | newline_list NEWLINE'''
    pass

def p_single_arg(t):
    'single_arg : arg_type NAME arg_opt'
    comment, opt = t[3]
    t[0] = Node(type='arg', c_type=t[1], name=t[2], comment=comment, opt=opt)

def p_arg_type_1(t):
    'arg_type : NAME'
    t[0] = t[1]

def p_arg_type(t):
    'arg_type : NAME STAR'
    t[0] = t[1] + ' *'

def p_arg_opt_0(t):
    'arg_opt : '
    t[0] = '', None

def p_arg_opt_1(t):
    'arg_opt : COMMENT'
    t[0] = uncomment(t[1]), None

def p_arg_opt_2(t):
    'arg_opt : COLON COMMENT NEWLINE INDENT opt DEDENT'
    t[0] = uncomment(t[2]), t[5]

# List of callbacks statement
def p_callbacks_stmt(t):
    'callbacks_stmt : CALLBACKS COLON NEWLINE INDENT callback_list DEDENT'
    t[0] = Node(type='callbacklist', list=t[5])

def p_callback_list_0(t):
    'callback_list : callback_list NEWLINE'
    t[0] = t[1]

def p_callback_list_1(t):
    'callback_list : single_callback'
    t[0] = [t[1]]

def p_callback_list_2(t):
    'callback_list : callback_list single_callback'
    t[1].append(t[2])
    t[0] = t[1]

def p_single_callback_1(t):
    'single_callback : NAME NAME NEWLINE'
    t[0] = Node(type='callback', c_type=t[1], name=t[2], args=None)

def p_single_callback_2(t):
    'single_callback : NAME NAME COLON NEWLINE INDENT callback_args DEDENT'
    t[0] = Node(type='callback', c_type=t[1], name=t[2], args=t[6])

def p_callback_args_0(t):
    'callback_args : callback_args NEWLINE'
    t[0] = t[1]

def p_callback_args_1(t):
    'callback_args : single_arg'
    t[0] = [t[1]]

def p_callback_args_2(t):
    'callback_args : callback_args single_arg'
    t[1].append(t[2])
    t[0] = t[1]

def p_error(t):
    print "Syntax error at %s" % t

import yacc
yacc.yacc()

def parse(fp):
    return yacc.parse(lexer = Lexer(fp))
